class ApxException(IOError):
    """
    There was an ambiguous exception
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize WuxException
        """
        super().__init__(*args, **kwargs)

class MissingParam(ApxException):
    """For handling required param is missing"""

class UnSupported(ApxException):
    """For handling unsupported conditional"""

class InvalidValue(ApxException):
    """For handling if some value is invalid"""

class InvalidRequest(ApxException):
    """For handling invalid request"""

class BadReq(ApxException):
    """For handling bad request"""

class SpecificErrResp(ApxException):
    """For spessific error response"""



import typing as t
from apx.param import Param
from apx.config import ApxConfig
from apx.helpers import capute_err
from flask import current_app as cr_app
from apx.exceptions import (
    BadReq, 
    SpecificErrResp
)


def route(
    param:t.Optional[bool]=True, 
    allowed_role:t.Optional[t.Union[tuple, list]]=tuple()
):
    def decorator(func:t.Callable[[any, any], any]):
        def warper(*args, **kwargs):
            apx:ApxConfig = cr_app.apx
            try:
                if not apx.auth.is_authorized(allowed_role):
                    return {
                        "err": False, 
                        "msg": "Unauthorized", 
                        "data": None
                    }
                if param:
                    resp = func(param=Param(), *args, **kwargs)
                else:
                    resp = func(*args, **kwargs)
                return apx.resp.success(resp)
            except BadReq as e:
                return apx.resp.bad_req(e)
            except SpecificErrResp as e:
                return apx.resp.sepesific_err_resp(e)
            except:
                return {
                    "err": True, 
                    "msg": "server error", 
                    "data": None, 
                    "err_data": capute_err() if cr_app.debug else dict()
                }
        warper.__name__ = func.__name__
        return warper
    return decorator




import typing as t
from flask import request
from apx.helpers import try_it
from apx.exceptions import (
    BadReq, UnSupported, InvalidValue
)


class Param:
    '''
    Parse param value on GET and POST method
    '''
    def __init__(self) -> None:
        self.params = dict()
        self.json = False
        self.content_type = None

    def missing_param(self, name:str):
        raise BadReq(f"param {name} is require")

    def invalid_param_value(self, name:str, msg:str):
        raise BadReq(f"param {name} must be {msg}")

    def dump(self, method:str):
        if method == "GET":
            self.params = dict(request.args).copy()
        elif method == "POST":
            if self.content_type is None:
                content_type = request.headers.get('Content-Type') or request.headers.get('content-type')
                content_type = str(content_type).lower()
                self.content_type = content_type

            if "application/json" in self.content_type:
                self.json = True
                self.params = request.json
            else:
                self.params = dict(request.form)
        else:
            raise UnSupported(f"Unsupported request method {method}")

    def text(
        self, 
        #: param name default None for getting full json data
        name:t.Optional[str]=None, 
        expected_type=...,  
        is_require:bool=False, 
        default_val=None
    ):
        """
        getting param value by param name in request GET or POST method
        """
        if default_val is not None and is_require:
            raise InvalidValue("default value can't set if is_require=True")

        if self.params == dict(): self.dump(request.method)

        data = default_val
        if name is not None:
            data = try_it(
                try_func=lambda: self.params[name], 
                except_funcs=lambda: self.missing_param(name=name) if is_require else None, 
                default=default_val
            )

        if not isinstance(expected_type, type(Ellipsis)):
            if not isinstance(data, expected_type):
                self.invalid_param_value(name, f'{expected_type.__name__} not {type(data).__name__}')

        return data




import typing as t
from flask import session


class Auth:
    def __init__(self, list_role:t.Union[list, tuple], get_role_info:t.Callable) -> None:
        self.list_role = list_role
        self.get_role_info = get_role_info

    def update(self, data) -> None:
        session.update(data)

    def delete(self, keys:t.Union[list, tuple]) -> None:
        tuple(session.pop(key, None) for key in keys)

    def get(self) -> dict:
        return dict(session)

    def is_authorized(self, allowed_role:t.Union[list, tuple]) -> bool:
        if len(allowed_role) == 0: return True
        return self.get_role_info(self.get()) in allowed_role



import sys
import traceback
import typing as t
from flask import current_app as cr_app


class ApxResp:
    def __init__(self) -> None:
        self.spesific_err_msgs = tuple()

    def success(self, data):
        return {
            "err": False, 
            "msg": "success", 
            "data": data
        }

    def bad_req(self, e):
        return {
            "err": True, 
            "msg": str(e), 
            "data": None
        }

    def sepesific_err_resp(self, e):
        return {
            "err": True, 
            "msg": parse_err_msg(self.spesific_err_msgs, e) if cr_app.debug else "", 
            "data": None
        }


def try_it(
    #: The function will be call in try statment
    try_func, 
    #: function will be call in except statment, assign dict for spessific exception
    except_funcs:t.Optional[t.Union[t.Callable, dict]]=None,  
    #: the data will be return in except statment
    default=None
):
    try:
        return try_func()
    except Exception as e:
        if except_funcs is not None:
            if isinstance(except_funcs, dict):
                except_func = except_funcs.get(type(e))
                if except_func is None: except_func = except_funcs.get(Exception)
            else:
                except_func = except_funcs

            except_func()
        return default


def capute_err() -> dict:
    ex_type, ex_value, ex_traceback = sys.exc_info()
    return {
        "exception": f"\n{ex_type.__name__}: {ex_value}", 
        "stack":     [
            f"File {trace[0]}, Line {trace[1]}, in {trace[2]}\n{trace[3]}" 
            for trace in traceback.extract_tb(ex_traceback)
        ]
    }


def parse_err_msg(err_list:t.Union[list, tuple], code:str):
    code = str(code).split(":")
    data = tuple()
    if len(code) > 1: data = tuple(code[1:])
    try: return f"{code[0]} >> {err_list[int(code[0])].format(*data)}"
    except: return f"Unknown code: {code}"



from setuptools import setup


with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()

with open("README.rst") as f:
    readme = f.read()

github_url = "https://github.com/Nux-xader/apx"


setup(
    name="FlaskApx", 
    url=github_url, 
    project_urls={
        "Github": github_url
    },
    author="Nux xader A.K.A Satria Rahmat", 
    python_requires=">=3.7, <4", 
    version="1.0.0", 
    keywords=["flask helper", "flask", "API heler", "API Toolkit", "API warper"], 
    description="The helper for Flask", 
    long_description=readme, 
    packages=["apx"], 
    install_requires=requirements
)
